/*
 * SPDX-FileCopyrightText: 2014-2020 Bruce Cowan <bruce@bcowan.me.uk>
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include <webkit2/webkit2.h>

#include "glob-app-window.h"
#include "glob-match-item.h"
#include "glob-match-widget.h"
#include "glob-search.h"

struct _GlobAppWindow
{
    GtkApplicationWindow parent_instance;

    GSettings *settings;
    GlobSearch *search;

    GtkWidget *header;
    GtkWidget *entry;
    GtkWidget *listbox;
    GtkWidget *popover;
    GtkWidget *error_label;
    GtkWidget *webview;
    GtkWidget *main_box;

    GCancellable *cancellable;
};

G_DEFINE_TYPE (GlobAppWindow, glob_app_window, GTK_TYPE_APPLICATION_WINDOW)

static void
row_activated_cb (G_GNUC_UNUSED GtkListBox    *box,
                                GtkListBoxRow *row,
                                gpointer       user_data)
{
    GlobAppWindow *self = GLOB_APP_WINDOW (user_data);

    GlobMatchItem *item = glob_match_widget_get_item (GLOB_MATCH_WIDGET (row));
    const char *text = glob_match_item_get_word (item);

    g_autofree char *uri = g_strdup_printf ("https://en.m.wiktionary.org/wiki/%s", text);
    webkit_web_view_load_uri (WEBKIT_WEB_VIEW (self->webview), uri);
}

static GtkWidget *
create_widget (              gpointer item,
               G_GNUC_UNUSED gpointer user_data)
{
    return glob_match_widget_new (GLOB_MATCH_ITEM (item));
}

static void
populate_finish (GObject      *source_object,
                 GAsyncResult *res,
                 gpointer      user_data)
{
    GlobAppWindow *self = GLOB_APP_WINDOW (user_data);
    unsigned n;
    GError *err = NULL;

    g_autoptr (GListModel) list = glob_search_get_matches_finish (GLOB_SEARCH (source_object),
                                                                  res,
                                                                  &err);

    GtkStyleContext *entry_ctx = gtk_widget_get_style_context (self->entry);

    if (list)
    {
        gtk_style_context_remove_class (entry_ctx, "error");
        gtk_popover_popdown (GTK_POPOVER (self->popover));

        n = g_list_model_get_n_items (list);
    }
    else
    {
        gtk_style_context_add_class (entry_ctx, "error");
        gtk_label_set_text (GTK_LABEL (self->error_label), err->message);
        gtk_popover_popup (GTK_POPOVER (self->popover));

        n = 0;
    }

    gtk_list_box_bind_model (GTK_LIST_BOX (self->listbox),
                             list, create_widget, self, NULL);

    if (n == 0)
        gtk_header_bar_set_subtitle (GTK_HEADER_BAR (self->header), "Found no matches");
    else if (n == 1)
        gtk_header_bar_set_subtitle (GTK_HEADER_BAR (self->header), "Found 1 match");
    else
    {
        g_autofree char *status = g_strdup_printf ("Found %u matches", n);
        gtk_header_bar_set_subtitle (GTK_HEADER_BAR (self->header), status);
    }
}

static void
entry_search_changed_cb (GtkSearchEntry *entry,
                         GlobAppWindow  *self)
{
    g_cancellable_cancel (self->cancellable);
    g_clear_object (&self->cancellable);
    self->cancellable = g_cancellable_new ();

    if (gtk_entry_get_text_length (GTK_ENTRY (entry)) > 0)
    {
        const char *pattern = gtk_entry_get_text (GTK_ENTRY (entry));
        glob_search_get_matches (self->search, pattern,
                                 self->cancellable, populate_finish, self);
    }
}

static void
settings_changed_cb (G_GNUC_UNUSED GSettings *settings,
                                   char      *key,
                                   gpointer   user_data)
{
    if (g_strcmp0 (key, "mode") == 0)
        gtk_entry_set_text (GTK_ENTRY (GLOB_APP_WINDOW (user_data)->entry), "");
}

static gboolean
webview_context_menu_cb (G_GNUC_UNUSED WebKitWebView       *view,
                         G_GNUC_UNUSED WebKitContextMenu   *menu,
                         G_GNUC_UNUSED GdkEvent            *event,
                         G_GNUC_UNUSED WebKitHitTestResult *hit_test_result,
                         G_GNUC_UNUSED gpointer             user_data)
{
    return TRUE;
}

static gboolean
webview_decide_policy_cb (G_GNUC_UNUSED WebKitWebView           *view,
                                        WebKitPolicyDecision    *decision,
                                        WebKitPolicyDecisionType decision_type,
                          G_GNUC_UNUSED gpointer                 user_data)
{
    if (decision_type == WEBKIT_POLICY_DECISION_TYPE_NAVIGATION_ACTION)
    {
        WebKitNavigationPolicyDecision *nav_decision =
            WEBKIT_NAVIGATION_POLICY_DECISION (decision);
        WebKitNavigationAction *action =
            webkit_navigation_policy_decision_get_navigation_action (nav_decision);

        if (webkit_navigation_action_get_navigation_type (action) != WEBKIT_NAVIGATION_TYPE_OTHER)
        {
            webkit_policy_decision_ignore (decision);
            return TRUE;
        }
    }

    return FALSE;
}

static void
glob_app_window_dispose (GObject *object)
{
    GlobAppWindow *self = GLOB_APP_WINDOW (object);

    g_clear_object (&self->settings);

    G_OBJECT_CLASS (glob_app_window_parent_class)->dispose (object);
}

static void
glob_app_window_class_init (GlobAppWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    object_class->dispose = glob_app_window_dispose;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 "/uk/me/bcowan/wordsearch/main.ui");

    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, header);
    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, entry);
    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, listbox);
    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, popover);
    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, error_label);
    gtk_widget_class_bind_template_child (widget_class, GlobAppWindow, main_box);

    gtk_widget_class_bind_template_callback (widget_class, entry_search_changed_cb);
    gtk_widget_class_bind_template_callback (widget_class, row_activated_cb);
}

static void
glob_app_window_init (GlobAppWindow *self)
{
    gtk_widget_init_template (GTK_WIDGET (self));
    gtk_window_maximize (GTK_WINDOW (self));

    self->settings = g_settings_new ("uk.me.bcowan.Wordsearch");
    g_signal_connect (self->settings, "changed",
                      G_CALLBACK (settings_changed_cb), self);

    self->search = glob_search_new ();

    self->webview = webkit_web_view_new ();
    g_signal_connect (self->webview, "context-menu",
                      G_CALLBACK (webview_context_menu_cb), NULL);
    g_signal_connect (self->webview, "decide-policy",
                      G_CALLBACK (webview_decide_policy_cb), NULL);
    webkit_web_view_load_uri (WEBKIT_WEB_VIEW (self->webview),
                              "https://en.m.wiktionary.org/wiki/");

    gtk_box_pack_end (GTK_BOX (self->main_box), self->webview,
                      TRUE, TRUE, 0);
}

/**
 * glob_app_window_new:
 * @app: the application object
 *
 * Gets an application window with all its widgets setup for use.
 *
 * Returns: an application window
 */
GtkWidget *
glob_app_window_new (GtkApplication *app)
{
    return g_object_new (GLOB_TYPE_APP_WINDOW,
                         "application", app,
                         NULL);
}
