/*
 * SPDX-FileCopyrightText: 2016-2020 Bruce Cowan <bruce@bcowan.me.uk>
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "glob-match-item.h"

enum
{
    PROP_0,
    PROP_WORD,
    PROP_MATCH,
    PROP_LAST
};

static GParamSpec *pspecs[PROP_LAST];

struct _GlobMatchItem
{
    GObject parent_instance;

    char *word;
    GMatchInfo *match;
};

G_DEFINE_TYPE (GlobMatchItem, glob_match_item, G_TYPE_OBJECT)

static void
glob_match_item_get_property (GObject    *object,
                              unsigned    property_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
    GlobMatchItem *self = GLOB_MATCH_ITEM (object);

    switch (property_id)
    {
    case PROP_WORD:
        g_value_set_string (value, glob_match_item_get_word (self));
        break;
    case PROP_MATCH:
        g_value_set_boxed (value, glob_match_item_get_match (self));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
glob_match_item_set_property (GObject      *object,
                              unsigned      property_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
    GlobMatchItem *self = GLOB_MATCH_ITEM (object);

    switch (property_id)
    {
    case PROP_WORD:
        self->word = g_value_dup_string (value);
        break;
    case PROP_MATCH:
        self->match = g_value_dup_boxed (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
glob_match_item_finalize (GObject *object)
{
    GlobMatchItem *self = GLOB_MATCH_ITEM (object);

    g_clear_pointer (&self->word, g_free);

    G_OBJECT_CLASS (glob_match_item_parent_class)->finalize (object);
}

static void
glob_match_item_dispose (GObject *object)
{
    GlobMatchItem *self = GLOB_MATCH_ITEM (object);

    g_clear_pointer (&self->match, g_match_info_unref);

    G_OBJECT_CLASS (glob_match_item_parent_class)->dispose (object);
}

static void
glob_match_item_class_init (GlobMatchItemClass *klass)
{
    GObjectClass *obj_class = G_OBJECT_CLASS (klass);

    obj_class->get_property = glob_match_item_get_property;
    obj_class->set_property = glob_match_item_set_property;
    obj_class->finalize = glob_match_item_finalize;
    obj_class->dispose = glob_match_item_dispose;

    pspecs[PROP_WORD] = g_param_spec_string ("word", "Word", "Dictionary word",
                                             "",
                                             G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    pspecs[PROP_MATCH] = g_param_spec_boxed ("match", "Match", "Regex match",
                                             G_TYPE_MATCH_INFO,
                                             G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_properties (obj_class, PROP_LAST, pspecs);
}

static void
glob_match_item_init (G_GNUC_UNUSED GlobMatchItem *self)
{

}

/**
 * glob_match_item_new:
 *
 * Create a new #GlobMatchItem, which represents a match.
 *
 * Returns: a new #GlobMatchItem
 */
GlobMatchItem *
glob_match_item_new (const char  *word,
                     GMatchInfo  *match)
{
    return g_object_new (GLOB_TYPE_MATCH_ITEM,
                         "word", word,
                         "match", match,
                         NULL);
}

/**
 * glob_match_item_get_word:
 * @self: a #GlobMatchItem
 *
 * Gets the word represented by @self.
 *
 * Returns: (transfer none): the word
 */
const char *
glob_match_item_get_word (GlobMatchItem *self)
{
    g_assert (GLOB_IS_MATCH_ITEM (self));

    return self->word;
}

/**
 * glob_match_item_get_match:
 * @self: a #GlobMatchItem
 *
 * Gets the regex match object associated associated with the match item.
 *
 * Returns: (transfer none): the match object
 */
GMatchInfo *
glob_match_item_get_match (GlobMatchItem *self)
{
    g_assert (GLOB_IS_MATCH_ITEM (self));

    return self->match;
}
