/*
 * SPDX-FileCopyrightText: 2012-2020 Bruce Cowan <bruce@bcowan.me.uk>
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "glob-match-item.h"
#include "glob-search.h"

G_DEFINE_QUARK (GlobSearchError, glob_search_error)

typedef enum
{
    MATCH_MODE_GLOB = 1,
    MATCH_MODE_REGEX = 2,
} MatchMode;

struct _GlobSearch
{
    GObject parent_instance;

    GSettings *settings;
};

G_DEFINE_TYPE (GlobSearch, glob_search, G_TYPE_OBJECT)

static void
glob_search_dispose (GObject *object)
{
    GlobSearch *self = GLOB_SEARCH (object);

    g_clear_object (&self->settings);

    G_OBJECT_CLASS (glob_search_parent_class)->dispose (object);
}

static void
glob_search_class_init (GlobSearchClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = glob_search_dispose;
}

static void
glob_search_init (GlobSearch *self)
{
    self->settings = g_settings_new ("uk.me.bcowan.Wordsearch");
}

static GListModel *
get_matches (G_GNUC_UNUSED GlobSearch    *self,
                           const char    *pattern,
                           GCancellable  *cancellable,
                           GError       **err)
{
    g_autoptr(GRegex) regex = NULL;
    g_autoptr(GPatternSpec) pspec = NULL;

    int mode = g_settings_get_enum (self->settings, "mode");

    if (mode == MATCH_MODE_REGEX)
    {
        regex = g_regex_new (pattern, G_REGEX_CASELESS, 0, err);

        if (regex == NULL)
            return NULL;
    }
    else if (mode == MATCH_MODE_GLOB)
    {
        pspec = g_pattern_spec_new (pattern);
    }
    else
    {
        g_assert_not_reached ();
    }

    // Set up data input stream
    g_autoptr(GInputStream) is = g_resources_open_stream ("/uk/me/bcowan/wordsearch/words.txt",
                                                          G_RESOURCE_LOOKUP_FLAGS_NONE,
                                                          err);
    if (is == NULL)
        return NULL;

    g_autoptr(GDataInputStream) dis = g_data_input_stream_new (is);

    // Use a temporary pointer array, and add all entries to list model at once
    g_autoptr(GPtrArray) tmp = g_ptr_array_new_with_free_func (g_object_unref);

    while (TRUE)
    {
        size_t length;
        g_autofree char *line = g_data_input_stream_read_line_utf8 (dis, &length,
                                                                    cancellable,
                                                                    err);
        if (G_UNLIKELY (line == NULL))  // End of file
            break;

        if (mode == MATCH_MODE_REGEX)
        {
            g_autoptr(GMatchInfo) match;

            if (G_UNLIKELY (g_regex_match (regex, line, 0, &match)))
            {
                GlobMatchItem *item = glob_match_item_new (line, match);
                g_ptr_array_add (tmp, item);
            }
        }
        else if (mode == MATCH_MODE_GLOB)
        {
            if (G_UNLIKELY (g_pattern_match (pspec, length, line, NULL)))
            {
                GlobMatchItem *item = glob_match_item_new (line, NULL);
                g_ptr_array_add (tmp, item);
            }
        }

        // Rate limit
        if (G_UNLIKELY (tmp->len == 1000))
        {
            g_set_error (err,
                         GLOB_SEARCH_ERROR,
                         GLOB_SEARCH_TOO_MANY_RESULTS,
                         "Too many results to display");
            return NULL;
        }
    }

    GListStore *store = g_list_store_new (GLOB_TYPE_MATCH_ITEM);
    g_list_store_splice (store, 0, 0, tmp->pdata, tmp->len);

    return G_LIST_MODEL (store);
}

static void
get_matches_async_thread (GTask        *task,
                          gpointer      source_object,
                          gpointer      task_data,
                          GCancellable *cancellable)
{
    GError *err = NULL;

    char *pattern = (char *) task_data;
    GListModel *matches = get_matches (GLOB_SEARCH (source_object), pattern,
                                       cancellable, &err);

    if (err == NULL)
        g_task_return_pointer (task, matches, (GDestroyNotify) g_object_unref);
    else
        g_task_return_error (task, err);
}

/**
 * glob_search_get_matches:
 * @self: a #GlobSearch
 * @pattern: the pattern to match against
 * @cancellable: (allow-none): optional #GCancellable object, %NULL to ignore
 * @callback: (scope async): the callback to run when finished
 * @user_data: (closure): data to pass to the callback
 *
 * Gets the list of words that @glob matches against asynchronously.
 * When this finishes, @callback will be called. Call
 * glob_search_get_matches_finish() to get the result.
 */
void
glob_search_get_matches (GlobSearch          *self,
                         const char          *pattern,
                         GCancellable        *cancellable,
                         GAsyncReadyCallback  callback,
                         gpointer             user_data)
{
    g_assert (GLOB_IS_SEARCH (self));
    g_assert (pattern != NULL);
    g_assert (G_IS_CANCELLABLE (cancellable) || cancellable == NULL);
    g_assert (callback != NULL);

    char *data = g_strdup (pattern);
    g_autoptr(GTask) task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_task_data (task, data, g_free);
    g_task_run_in_thread (task, get_matches_async_thread);
}

/**
 * glob_search_get_matches_finish:
 * @self: a #GlobSearch.
 * @result: a #GAsyncResult
 * @error: a #GError, %NULL to ignore
 * 
 * Finishes an asynchronous pattern matching operation.
 *
 * Returns: (transfer full): the list of matches, free with g_object_unref()
 *     when finished
 */
GListModel *
glob_search_get_matches_finish (GlobSearch    *self,
                                GAsyncResult  *res,
                                GError       **error)
{
    g_assert (GLOB_IS_SEARCH (self));
    g_assert (G_IS_ASYNC_RESULT (res));
    g_assert (g_task_is_valid (res, self));
    g_assert (error == NULL || *error == NULL);

    return g_task_propagate_pointer (G_TASK (res), error);
}

/**
 * glob_search_new:
 *
 * Create a new #GlobSearch. This is used to search the word list.
 *
 * Returns: a new #GlobSearch
 */
GlobSearch *
glob_search_new (void)
{
    return g_object_new (GLOB_TYPE_SEARCH, NULL);
}
